-- Copyright (C) 2012 Jared Krinke.
-- 
-- This is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

require("UI/Input.lua")
require("UI/Label.lua")
require("UI/Button.lua")
require("UI/Form.lua")

-- Helper layer for setting a new control
local rawEventCaptureLayer = Layer.new(-1)

rawEventCaptureLayer:addChild(Entity.new(0, 0, 0, Form.defaultTextWidth, Form.defaultTextHeight, 0, Color.new(1, 1, 1), ElementList.new(Element.Text.new("Press any key", 0, 0, 0, 1, 1, 0, Color.new(1, 1, 1), "center"))))

-- TODO: Add some UI for binding (preferably without hiding lower layers)
function rawEventCaptureLayer.rawEventReceived(self, rawEvent, pressed)
    if isString(rawEvent) then
        if pressed then
            self.rawEvent = rawEvent
        elseif not pressed and not isNil(self.rawEvent) then
            Layers.pop()
        
            if isFunction(self.rawEventCaptured) then
                self.rawEventCaptured()
            end
        end
    end
end

function rawEventCaptureLayer.keyPressed(self, key, pressed, character)
    self:rawEventReceived(key, pressed)
end

function rawEventCaptureLayer.mouseButtonPressed(self, button, pressed, x, y)
    self:rawEventReceived(button, pressed)
end

function rawEventCaptureLayer.joystickButtonPressed(self, joystick, button, pressed)
    self:rawEventReceived(Input.translateJoystickButtonToRawEvent(joystick, button), pressed)
end

function rawEventCaptureLayer.joystickAxisMoved(self, joystick, axis, value)
    local rawEventPressed
    local rawEventReleased

    rawEventPressed, rawEventReleased = Input.translateJoystickAxisMoveToRawEvent(self, joystick, axis, value)

    self:rawEventReceived(rawEventPressed, true)
    self:rawEventReceived(rawEventReleased, false)
end

function rawEventCaptureLayer.capture(self, rawEventCaptured)
    self.rawEvent = nil
    self.rawEventCaptured = rawEventCaptured
    Layers.push(self)
end

function rawEventCaptureLayer.getRawEvent(self)
    return self.rawEvent
end

local function writeRawEventsToStringBuffer(rawEvents, stringBuffer)
    local first = true

    stringBuffer:clear()

    if isTable(rawEvents) then
        rawEvents:forEach(function(rawEvent)
            if first then
                first = false
            else
                stringBuffer:append(" ")
            end

            stringBuffer:append(rawEvent)
        end)
    end
end

-- The actual control component
Control = { }

function Control.new(input, logicalEvent, name)
    local labelText = logicalEvent

    if isString(name) then
        labelText = name
    end

    local rawEvents = Label.new("", "right")
    local control = Form.newNestedBox({ left = Button.new(labelText), center = rawEvents })

    control.input = input
    control.logicalEvent = logicalEvent
    control.rawEvents = rawEvents
    control.buffer = StringBuffer.new()

    function control.notified(self)
        writeRawEventsToStringBuffer(self.input:getRawEvents(self.logicalEvent), self.buffer)
        self.rawEvents:setText(self.buffer.text)
        self:layout()
    end

    function control.activated(self)
        rawEventCaptureLayer:capture(function()
            self.input:bind(rawEventCaptureLayer:getRawEvent(), self.logicalEvent)
        end)
    end

    -- Treat as one component
    control.mouseMoved = nil

    -- Override inputReceived
    control.inputReceivedBase = control.inputReceived

    function control.inputReceived(self, event, pressed)
        local handled = false

        if event == "delete" and pressed then
            -- Remove oldest binding
            local rawEvents = self.input:getRawEvents(self.logicalEvent)

            if rawEvents:getCount() > 0 then
                self.input:unbind(rawEvents:get(1), self.logicalEvent)
            end

            handled = true
        end

        if not handled then
            handled = self:inputReceivedBase(event, pressed)
        end

        return handled
    end

    input:addListener(function() control:notified() end)
    control:notified()

    return control
end

