-- Copyright (C) 2012 Jared Krinke.
-- 
-- This is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

require("Utility/ForEach.lua")
require("Utility/LinkedList.lua")

-- TODO: There should be a way to layout components within a single form

Form = {
    defaultTextWidth = 15,
    defaultTextHeight = 30,
    defaultX = -200,
    defaultY = 200,
    defaultWidth = 400,
}

local locations = { "top", "left", "right", "center", "bottom" }

function Form.setDefaults(x, y, desiredWidth)
    Form.defaultX = x
    Form.defaultY = y
    Form.defaultWidth = desiredWidth
end

function Form.setDefaultTextSize(width, height)
    Form.defaultTextWidth = width
    Form.defaultTextHeight = height

    -- TODO: Should this update existing forms? Probably not...
end

local function gridFlowLayout(self, pad, alignment)
    local columns = self.columns
    
    if not isNumber(columns) then
        columns = 1
    end
    
    -- Determine minimum column widths
    local columnWidths = { }
    local rowMinimumWidths = { 0 }
    local rowHeights = { 0 }
    local column = 1
    local row = 1
    
    while column <= columns do
        columnWidths[column] = 0
        column = column + 1
    end
    
    column = 1
    self.components:forEach(function(component)
        local minimumWidth, minimumHeight = component:getMinimumSize()
    
        columnWidths[column] = max(columnWidths[column], minimumWidth)
        rowMinimumWidths[row] = rowMinimumWidths[row] + minimumWidth
        rowHeights[row] = max(rowHeights[row], minimumHeight)
    
        if (column % columns) == 0 then
            row = row + 1
            rowMinimumWidths[row] = 0
            rowHeights[row] = 0
        end
    
        column = (column % columns) + 1
    end)
    
    -- Calculate total width
    local minimumWidth = 0
    
    column = 1
    while column <= columns do
        minimumWidth = minimumWidth + columnWidths[column]
        column = column + 1
    end
    
    self.minimumWidth = minimumWidth
    
    -- Check for size constraints
    if isNumber(self.desiredWidth) then
        -- Add padding (or negative padding)
        local padding = (self.desiredWidth - minimumWidth) / columns

        if not pad and padding > 0 then
            padding = 0
        end
    
        column = 1
        while column <= columns do
            -- Add padding
            columnWidths[column] = columnWidths[column] + padding
            column = column + 1
        end
    end
    
    -- Layout components
    local x = self.x
    local y = self.y
    local rowHeight = 0
    local totalMinimumHeight = 0
    
    column = 0
    row = 0
    
    self.components:forEach(function(component)
        if column == 0 or (column % columns) == 0 then
            -- Last column; move to next row
            if row > 0 then
                y = y - rowHeights[row]
                totalMinimumHeight = totalMinimumHeight + rowHeights[row]
            end

            x = self.x
    
            column = 1
            row = row + 1

            -- Special case for centered flow layout
            if not pad and isNumber(self.desiredWidth) and alignment == "center" then
                x = self.x + self.desiredWidth / 2 - rowMinimumWidths[row] / 2
            end
        else
            column = column + 1
        end
    
        local componentWidth = columnWidths[column]

        component:setPosition(x, y)

        -- Flow layout doesn't fill the entire column
        if not pad then
            componentWidth = component:getMinimumSize()
        end

        component:setSize(componentWidth, rowHeights[row])
        x = x + componentWidth
    end)
    
    if isNumber(rowHeights[row]) then
        totalMinimumHeight = totalMinimumHeight + rowHeights[row]
    end
    
    self.minimumHeight = totalMinimumHeight
end

local layouts = {
    grid = function(self)
        return gridFlowLayout(self, true)
    end,

    flow = function(self)
        return gridFlowLayout(self, false, self.alignment)
    end,

    box = function(self)
        local top = self.locationToComponent.top
        local left = self.locationToComponent.left
        local right = self.locationToComponent.right
        local center = self.locationToComponent.center
        local bottom = self.locationToComponent.bottom

        local topMinimumWidth, topMinimumHeight
        local leftMinimumWidth, leftMinimumHeight
        local rightMinimumWidth, rightMinimumHeight
        local centerMinimumWidth, centerMinimumHeight
        local bottomMinimumWidth, bottomMinimumHeight

        if not isNil(top) then
            topMinimumWidth, topMinimumHeight = top:getMinimumSize()
        end

        if not isNil(left) then
            leftMinimumWidth, leftMinimumHeight = left:getMinimumSize()
        end

        if not isNil(right) then
            rightMinimumWidth, rightMinimumHeight = right:getMinimumSize()
        end

        if not isNil(center) then
            centerMinimumWidth, centerMinimumHeight = center:getMinimumSize()
        end

        if not isNil(bottom) then
            bottomMinimumWidth, bottomMinimumHeight = bottom:getMinimumSize()
        end

        -- Calculate minimum width
        local minimumWidth = 0

        if not isNil(left) then
            minimumWidth = minimumWidth + leftMinimumWidth
        end

        if not isNil(center) then
            minimumWidth = minimumWidth + centerMinimumWidth
        end

        if not isNil(right) then
            minimumWidth = minimumWidth + rightMinimumWidth
        end

        if not isNil(top) and topMinimumWidth > minimumWidth then
            minimumWidth = topMinimumWidth
        end

        if not isNil(bottom) and bottomMinimumWidth > minimumWidth then
            minimumWidth = bottomMinimumWidth
        end

        self.minimumWidth = minimumWidth

        -- Check for size constraints
        local totalWidth = minimumWidth
        local padding = 0

        if isNumber(self.desiredWidth) then
            padding = self.desiredWidth - minimumWidth
            totalWidth = self.desiredWidth
        end

        -- Calculate middle row height
        local rowHeight = 0

        if not isNil(left) then
            rowHeight = max(rowHeight, leftMinimumHeight)
        end

        if not isNil(center) then
            rowHeight = max(rowHeight, centerMinimumHeight)
        end

        if not isNil(right) then
            rowHeight = max(rowHeight, rightMinimumHeight)
        end

        -- Layout components
        local x = self.x
        local y = self.y

        if not isNil(top) then
            top:setPosition(x, y)
            top:setSize(totalWidth, topMinimumHeight)
            y = y - topMinimumHeight
        end

        if not isNil(left) then
            left:setPosition(x, y)
            left:setSize(leftMinimumWidth, rowHeight)
            x = x + leftMinimumWidth
        end

        if not isNil(center) then
            center:setPosition(x, y)
            center:setSize(centerMinimumWidth + padding, rowHeight)
            x = x + centerMinimumWidth + padding
        end

        if not isNil(right) then
            right:setPosition(x, y)
            right:setSize(rightMinimumWidth, rowHeight)
        end
        
        x = self.x
        y = y - rowHeight

        if not isNil(bottom) then
            bottom:setPosition(x, y)
            bottom:setSize(totalWidth, bottomMinimumHeight)
            y = y - topMinimumHeight
        end

        self.minimumHeight = self.y - y
    end,
}

local setLayers = {
    grid = function(self, layer)
        self.components:forEach(function(component) component:setLayer(layer) end)
    end,

    flow = function(self, layer)
        self.components:forEach(function(component) component:setLayer(layer) end)
    end,

    box = function(self, layer)
        local i = 1

        while i <= #locations do
            local component = self.locationToComponent[locations[i]]

            if not isNil(component) then
                component:setLayer(layer)
            end

            i = i + 1
        end
    end,
}

function Form.new(layoutName, x, y, desiredWidth, desiredHeight)
    local form = { }

    form.components = LinkedList.new()
    form.count = 0
    form.layoutName = layoutName
    form.textWidth = Form.defaultTextWidth
    form.textHeight = Form.defaultTextHeight
    form.locationToComponent = { }

    -- These values handled nested forms (the latter two Booleans indicate if
    -- focus is outside this form
    form.nested = false

    if isNumber(x) then
        form.x = x
    else
        form.x = Form.defaultX
    end

    if isNumber(y) then
        form.y = y
    else
        form.y = Form.defaultY
    end

    if isNumber(desiredWidth) then
        form.desiredWidth = desiredWidth
    else
        form.desiredWidth = Form.defaultWidth
    end

    if isNumber(desiredHeight) then
        form.desiredHeight = desiredHeight
    end

    function form.changeFocus(self, newlyFocusedNode, event)
        local lastFocusedNode = self.focusedNode
        
        self.focusedNode = newlyFocusedNode

        if self.layer ~= nil then
            self.layer:play("Sounds/Move.wav", 0.5)
        end
        
        -- Notify the previously-focused component that it has lost focus
        if isTable(lastFocusedNode) and not isNil(lastFocusedNode.value) then
            local unfocusedHandler = lastFocusedNode.value.unfocused
    
            if isFunction(unfocusedHandler) then
                unfocusedHandler(lastFocusedNode.value)
            end
        end
        
        -- Notify the newly-focused component that it has gained focus
        if isTable(newlyFocusedNode) and not isNil(newlyFocusedNode.value) then
            local focusedHandler = newlyFocusedNode.value.focused
    
            if isFunction(focusedHandler) then
                focusedHandler(newlyFocusedNode.value, event)
            end
        end
    end

    function form.focus(self, component)
        local node = self.components.head 

        -- Linear search
        while isTable(node) and not isNil(node.value) do
            if node.value == component then
                self:changeFocus(node)
                break
            end

            node = node.next
        end
    end

    
    function form.moveFocusUp(self, event)
        local oldNode = self.focusedNode
        local newNode = nil

        if isTable(oldNode) then
            newNode = self.focusedNode.previous
        else
            -- Start with the last component
            newNode = self.components.head

            while isTable(newNode) and isTable(newNode.next) do
                newNode = newNode.next
            end
        end

        while isTable(newNode) and not isNil(newNode.value) and isFunction(newNode.value.getActive) and not newNode.value:getActive() do
            newNode = newNode.previous
        end

        if (isTable(newNode) and not isNil(newNode.value)) then
            self:changeFocus(newNode, event)
        end

        return not isNil(newNode) and newNode ~= oldNode
    end
    
    function form.moveFocusDown(self, event)
        local oldNode = self.focusedNode
        local newNode = nil

        if isTable(oldNode) then
            newNode = self.focusedNode.next
        else
            -- Start with the first component
            newNode = self.components.head
        end

        while isTable(newNode) and not isNil(newNode.value) and isFunction(newNode.value.getActive) and not newNode.value:getActive() do
            newNode = newNode.next
        end

        if (isTable(newNode) and not isNil(newNode.value)) then
            self:changeFocus(newNode, event)
        end

        return not isNil(newNode) and newNode ~= oldNode
    end
    
    function form.moveLeft(self)
        local handled = false

        if isTable(self.focusedNode)
            and not isNil(self.focusedNode.value)
            and isFunction(self.focusedNode.value.movedLeft)
        then
            handled = self.focusedNode.value:movedLeft()
        end

        return handled
    end
    
    function form.moveRight(self)
        local handled = false

        if isTable(self.focusedNode)
            and not isNil(self.focusedNode.value)
            and isFunction(self.focusedNode.value.movedRight)
        then
            handled = self.focusedNode.value:movedRight()
        end

        return handled
    end
    
    function form.activate(self)
        local handled = false

        if isTable(self.focusedNode)
            and not isNil(self.focusedNode.value)
            and isFunction(self.focusedNode.value.activated)
        then
            self.focusedNode.value:activated()
            handled = true
        end

        return handled
    end

    function form.keyPressed(self, key, pressed, character)
        local handled = false

        if isTable(self.focusedNode)
            and not isNil(self.focusedNode.value)
            and isFunction(self.focusedNode.value.keyPressed)
        then
            handled = self.focusedNode.value:keyPressed(key, pressed, character)
        end

        return handled
    end

    function form.mouseMoved(self, x, y)
        local handled = false
        local node = self.components.head 

        -- Check for active components that contain the mouse
        while isTable(node) and not isNil(node.value) do
            local component = node.value

            if not isFunction(component.getActive) or component:getActive() then
                local x1, y1 = component:getPosition()
                local width, height = component:getSize()
                local x2 = x1 + width
                local y2 = y1 - height

                if x >= x1 and x <= x2 and y <= y1 and y >= y2 then
                    local mouseMovedHandler = component.mouseMoved

                    if self.focusedNode ~= node then
                        self:changeFocus(node)
                    elseif isFunction(mouseMovedHandler) then
                        handled = mouseMovedHandler(component, x, y)
                    end

                    -- Only focus/notify the first match
                    break
                end
            end

            node = node.next
        end
    end
    
    function form.inputReceived(self, event, pressed)
        local eventHandler = self.eventHandlers[event]
        local handled = false
    
        -- Give priority to the focused component
        if isTable(self.focusedNode)
            and not isNil(self.focusedNode.value)
            and isFunction(self.focusedNode.value.inputReceived)
        then
            handled = self.focusedNode.value:inputReceived(event, pressed)
        end

        if not handled and isFunction(eventHandler) then
            if pressed then
                handled = eventHandler(self, event)
            end
        end

        return handled
    end
    
    function form.mouseButtonPressed(self, button, pressed)
        -- TODO
    end

    function form.layout(self)
        local layout = layouts[self.layoutName]

        if isFunction(layout) then
            layout(self)
        end

        if not isNil(self.parent) then
            self.parent:layout()
        end
    end
    
    function form.addAtLocation(self, location, component)
        local newNode = self.components:append(component)

        if isFunction(component.setTextSize) then
            component:setTextSize(self.textWidth, self.textHeight)
        end
        -- TODO: Should this be added to a layer

        -- If this is a nested form, set the "nested" property to true so focus
        -- changes outside this form are handled properly
        if isFunction(component.setParent) then
            component:setParent(self)
        end

        if isString(location) and self.layoutName == "box" then
            self.locationToComponent[location] = component
        end
    
        self.count = self.count + 1

        self:layout()

        return component
    end

    function form.add(self, component)
        local c = nil

        if self.layoutName == "grid" or self.layoutName == "flow" then
            c = self:addAtLocation(nil, component)
        end

        return c
    end

    function form.setLayer(self, layer)
        local setLayer = setLayers[self.layoutName]

        self.layer = layer

        if isFunction(setLayer) then
            setLayer(self, layer)
        end
    end

    function form.setParent(self, parent)
        self.parent = parent
        self.nested = true
    end

    -- Component implementation
    function form.getPosition(self)
        return self.x, self.y
    end

    function form.setPosition(self, x, y)
        if x ~= self.x or y ~= self.y then
            self.x = x
            self.y = y
            self:layout()
        end
    end

    function form.setTextSize(self, textWidth, textHeight)
        self.components:forEach(function(component)
            if isFunction(component.setTextSize) then
                component:setTextSize(textWidth, textHeight)
            end
        end)
    end

    function form.getActive(self)
        -- Check for at least one active component
        local node = self.components.head
        local active = false

        while isTable(node) and not isNil(node.value) do
            if isFunction(node.value.getActive) and not node.value:getActive() then
                node = node.next
            else
                active = true
                break
            end
        end

        return active
    end

    function form.getMinimumSize(self)
        if not isNumber(self.minimumWidth) or not isNumber(self.minimumHeight) then
            self:layout()
        end

        return self.minimumWidth, self.minimumHeight
    end

    function form.getDesiredSize(self)
        local width, height
        local minimumWidth, minimumHeight

        if not isNumber(self.desiredWidth) or not isNumber(self.desiredHeight) then
            minimumWidth, minimumHeight = self:getMinimumSize()
        end

        if isNumber(self.desiredWidth) then
            width = self.desiredWidth
        else
            width = minimumWidth
        end

        if isNumber(self.desiredHeight) then
            height = self.desiredHeight
        else
            height = minimumHeight
        end

        return width, height
    end

    function form.getSize(self)
        return self:getDesiredSize()
    end

    function form.setSize(self, width, height)
        -- TODO: What to do about height? Should that work like width?
        -- TODO: What if the size doesn't change but a re-layout is still required?
        if isNil(self.desiredWidth) or self.desiredWidth ~= width then
            self.desiredWidth = width
            self:layout()
        end
    end

    -- Nested forms can receive/lose focus (meaning a transition in/out of the form)
    function form.focused(self, event)
        if not isString(event) then
            -- Default to entering the form from the top
            self:moveFocusDown()
        else
            -- If an entry event was supplied, use that instead
            local eventHandler = self.eventHandlers[event]

            if isFunction(eventHandler) then
                eventHandler(self)
            end
        end
    end

    function form.unfocused(self)
        self:changeFocus(nil)
    end

    form.eventHandlers = {
        up = form.moveFocusUp,
        down = form.moveFocusDown,
        left = form.moveLeft,
        right = form.moveRight,
        activate = form.activate,
    }

    return form
end

function Form.newGrid(columns, x, y, desiredWidth, desiredHeight, ...)
    local form = Form.new("grid", x, y, desiredWidth, desiredHeight)

    form.columns = columns

    -- Add all components
    forEach(function(component)
        form:add(component)
    end, ...)

    return form
end

function Form.newFlow(columns, alignment, x, y, desiredWidth, desiredHeight, ...)
    local form = Form.new("flow", x, y, desiredWidth, desiredHeight)

    form.columns = columns
    form.alignment = alignment

    -- Add all components
    forEach(function(component)
        form:add(component)
    end, ...)

    return form
end

function Form.newLeftFlow(columns, x, y, desiredWidth, desiredHeight, ...)
    return Form.newFlow(columns, "left", x, y, desiredWidth, desiredHeight, ...)
end

function Form.newCenterFlow(columns, x, y, desiredWidth, desiredHeight, ...)
    return Form.newFlow(columns, "center", x, y, desiredWidth, desiredHeight, ...)
end

function Form.newNestedGrid(columns, ...)
    return Form.newGrid(columns, 0, 0, nil, nil, ...)
end

function Form.newNestedFlow(columns, ...)
    return Form.newFlow(columns, "left", 0, 0, nil, nil, ...)
end

function Form.newNestedLeftFlow(columns, ...)
    return Form.newFlow(columns, "left", 0, 0, nil, nil, ...)
end

function Form.newNestedCenterFlow(columns, ...)
    return Form.newFlow(columns, "center", 0, 0, nil, nil, ...)
end

function Form.newBox(x, y, desiredWidth, desiredHeight, locationToComponent)
    local form = Form.new("box", x, y, desiredWidth, desiredHeight)

    -- Add all components
    local i = 1

    while i <= #locations do
        local component = locationToComponent[locations[i]]

        if not isNil(component) then
            form:addAtLocation(locations[i], component)
        end

        i = i + 1
    end

    return form
end

function Form.newNestedBox(locationToComponent)
    return Form.newBox(0, 0, nil, nil, locationToComponent)
end

