-- Copyright (C) 2012 Jared Krinke.
-- 
-- This is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

require("UI/Input.lua")
require("Settings/Controls.lua")
require("UI/FormLayer.lua")
require("Effects/Ghost.lua")

-- Constants
local title = "Cretan Hubris"
local version = "0.1"

local textWidth = 15
local textHeight = 30

local directions = { { -1, 0 }, { 1, 0 }, { 0, 1 }, { 0, -1 } }
local directionValues = { left = 1, right = 2, up = 3, down = 4 }

-- Controls
local logicalEvents = {
    up = { necessary = true },
    down = { necessary = true },
    left = { necessary = true },
    right = { necessary = true },
    activate = { necessary = true },
    delete = { },
    cancel = { },
}

-- Set up some defaults
local input = Input.new(logicalEvents, controlsData)
FormLayer.setDefaultInput(input)

Form.setDefaults(-200, 200, 400)
Form.setDefaultTextSize(textWidth, textHeight)

-- Make necessary constants visible globally
CretanHubris = {
    title = title,
    version = version,
    input = input,
}

-- Classes
Vector = { }

function Vector.new(dimensions)
    local vector = { }
    local data = { }

    local function setRecursive(value, d, base, x, ...)
        local v = base[x]

        if d == 1 then
            base[x] = value
            return
        else
            if v == nil then
                v = { }
                base[x] = v
            end
        end

        return setRecursive(value, d - 1, v, ...)
    end

    function vector.set(self, value, ...)
        setRecursive(value, dimensions, data, ...)
    end

    local function getRecursive(d, base, x, ...)
        local v = base[x]

        if d == 1 then
            return v
        elseif v == nil then
            return nil
        end

        return getRecursive(d - 1, v, ...)
    end

    function vector.get(self, ...)
        return getRecursive(dimensions, data, ...)
    end

    local function removeRecursive(d, base, x, ...)
        local v = base[x]

        if d == 1 then
            base[x] = nil
            return
        elseif v == nil then
            return
        end

        return removeRecursive(d - 1, v, ...)
    end

    function vector.remove(self, ...)
        return removeRecursive(dimensions, data, ...)
    end

    function vector.clear(self)
        data = { }
    end

    return vector
end

VectorSet = { }

function VectorSet.new(dimensions)
    local vectorSet = { }
    local vector = Vector.new(dimensions)
    local list = { }
    local count = 0

    local function forEachArgument(f, x, ...)
        if x == nil then return end

        f(x)

        return forEachArgument(f, ...)
    end

    function vectorSet.add(self, ...)
        -- Add to set
        vector:set(true, ...)

        -- Add to list
        local item = { }

        forEachArgument(function(x)
            item[#item + 1] = x
        end, ...)

        list[#list + 1] = item

        -- Update count
        count = count + 1
    end

    local function tableToArgumentList(t, index)
        if index == #t then
            return t[index]
        else
            return t[index], tableToArgumentList(t, index + 1)
        end
    end

    function vectorSet.removeIndex(self, index)
        -- Remove from list (by swapping to the end and removing the last item)
        local item = list[index]

        if index ~= #list then
            list[index] = list[#list]
            list[#list] = item
        end

        list[#list] = nil

        -- Remove from set
        vector:remove(tableToArgumentList(item, 1))

        -- Update count
        count = count - 1

        return tableToArgumentList(item, 1)
    end

    function vectorSet.removeRandom(self)
        return self:removeIndex(random(count) + 1)
    end

    function vectorSet.contains(self, ...)
        return vector:get(...) ~= nil
    end

    function vectorSet.getCount(self)
        return count
    end

    function vectorSet.clear(self)
        vector:clear()
        list = { }
        count = 0
    end

    return vectorSet
end

Event = { }

function Event.new()
    local event = { }
    local handlers = { }

    function event.addHandler(self, f)
        handlers[#handlers + 1] = f
    end

    function event.signal(self, ...)
        local i = 1

        while i <= #handlers do
            handlers[i](...)
            i = i + 1
        end
    end

    return event
end

Player = {
    playerMoveChargePeriod = 200,
    playerMovePeriod = 150,
}

function Player.new(world)
    local player = Entity.new()

    local directionPressed = { }
    local moveTimer = 0
    local active = true

    -- Controls are handled in the follow priority order (most important to least)
    -- The intent is to allow right to be held while tweaking up/down
    local directionPriority = {
        directionValues.up,
        directionValues.down,
        directionValues.left,
        directionValues.right,
    }

    player.movedEvent = Event.new()

    function player.move(self, direction)
        if active then
            local offsets = directions[direction]

            if world:checkMove(self.x, self.y, direction) then
                self.x = self.x + offsets[1]
                self.y = self.y + offsets[2]
                world:ensureColumnsComplete(self.x)

                self.movedEvent:signal(self.x, self.y)

                return true
            end
        end

        return false
    end

    function player.checkAnyDirectionPressed(self)
        if active then
            local i = 1

            while i <= #directions do
                if directionPressed[i] == true then
                    return true
                end

                i = i + 1
            end
        end

        return false
    end

    function player.directionPressed(self, direction, pressed)
        if active and not self.checkAnyDirectionPressed() then
            self:move(direction)
            moveTimer = Player.playerMoveChargePeriod
        end

        directionPressed[direction] = pressed
    end

    function player.setActive(self, a)
        active = a
    end

    function player.update(self, ms)
        if active and self:checkAnyDirectionPressed() then
            moveTimer = moveTimer - ms

            while moveTimer <= 0 do
                local i = 1
                local moved = false

                -- Go through held controls in priority order
                while i <= #directionPriority and not moved do
                    local direction = directionPriority[i]

                    if directionPressed[direction] == true then
                        moved = self:move(direction)
                    end

                    i = i + 1
                end

                moveTimer = moveTimer + Player.playerMovePeriod
            end
        end
    end

    function player.reset(self)
        local i = 1

        while i <= #directions do
            directionPressed[i] = false
            i = i + 1
        end

        moveTimer = 0
        active = true

        player.x = 0
        player.y = 0
    end

    return player
end

Ender = { }

local enderInitialMovePeriod = 600

function Ender.new()
    local enderStartX = -14

    local ender = Entity.new(enderStartX)

    local movePeriod = enderInitialMovePeriod
    local moveTimer = movePeriod

    ender.movedEvent = Event.new()

    function ender.setMovePeriod(self, ms)
        movePeriod = ms
    end

    function ender.update(self, ms)
        moveTimer = moveTimer - ms

        while moveTimer <= 0 do
            self.x = self.x + 1
            self.movedEvent:signal(self.x)
            moveTimer = moveTimer + movePeriod
        end
    end

    function ender.reset(self)
        ender.x = enderStartX
        movePeriod = enderInitialMovePeriod
        moveTimer = movePeriod
    end

    return ender
end

World = {
}

function World.new()
    local world = Entity.new()

    -- Bounds (inclusive, note that there is no limit to the right)
    world.x1 = 0
    world.y1 = -4
    world.y2 = 4

    local worldMinColumns = 32
    world.lastColumn = 0

    -- World contents
    world.walls = Vector.new(3)
    world.squares = Vector.new(2)

    -- World state
    world.squaresUsed = VectorSet.new(2)
    world.wallsAvailable = VectorSet.new(3)
    world.wallsConsidered = VectorSet.new(3)

    -- Events
    world.changedEvent = Event.new()

    -- Constants
    local offsetWalls = { { -1, 0 }, { 0, 0 }, { 0, 0 }, { 0, -1 } }
    local wallAxes = { 2, 2, 1, 1 }
    local wallOffsets = {
        { { 0, 0 }, { 0, 1 } },
        { { 0, 0 }, { 1, 0 } },
    }

    function world.forEachSquareWall(self, x, y, f)
        local i = 1

        while i <= #offsetWalls do
            local offset = offsetWalls[i]
            local wallX = x + offset[1]
            local wallY = y + offset[2]

            f(wallAxes[i], wallX, wallY, i)

            i = i + 1
        end
    end

    function world.ensureWalls(self, x, y)
        self:forEachSquareWall(x, y, function(axis, x, y, direction)
            if self.walls:get(axis, x, y) == nil then
                self.walls:set(true, axis, x, y)
            end
        end)
    end

    function world.ensureSquare(self, x, y)
        if self.squares:get(x, y) == nil then
            self.squares:set(true, x, y)
            self:ensureWalls(x, y)
        end
    end
    
    function world.checkWallRemovable(self, axis, x, y)
        local x1 = self.x1 - 1
        local y1 = self.y1 - 1
        local y2 = self.y2

        if axis == 1 then
            return x >= x1 and y > y1 and y < y2
        else
            return x > x1 and y >= y1 and y <= y2
        end
    end

    function world.annexSquare(self, x, y)
        local i = 1

        self.squaresUsed:add(x, y)
        self:ensureSquare(x, y)
        self:forEachSquareWall(x, y, function(axis, wallX, wallY)
            -- Skip already-queued/considered
            if not self.wallsConsidered:contains(axis, wallX, wallY)
                and not self.wallsAvailable:contains(axis, wallX, wallY)
                and self:checkWallRemovable(axis, wallX, wallY)
            then
                self.wallsAvailable:add(axis, wallX, wallY)
            end
        end)
    end

    function world.removeWall(self, axis, x, y)
        self.walls:set(false, axis, x, y)
    end

    function world.carve(self)
        local carved = false

        while not carved and self.wallsAvailable:getCount() > 0 do
            local axis, wallX, wallY = self.wallsAvailable:removeRandom()
            local offsets = wallOffsets[axis]

            self.wallsConsidered:add(axis, wallX, wallY)
            
            local i = 1
            
            while i <= #offsets do
                local offset = offsets[i]
                local x = wallX + offset[1]
                local y = wallY + offset[2]
            
                if not self.squaresUsed:contains(x, y) then
                    self:removeWall(axis, wallX, wallY)
                    self:annexSquare(x, y)
                    carved = true
                end
            
                i = i + 1
            end
        end
    end

    function world.ensureWallConsidered(self, axis, wallX, wallY)
        while not self.wallsConsidered:contains(axis, wallX, wallY) do
            self:carve()
        end
    end

    function world.ensureColumnComplete(self, x)
        local y = self.y1

        while y <= self.y2 do
            self:ensureSquare(x, y)
            self:forEachSquareWall(x, y, function(axis, wallX, wallY)
                if self:checkWallRemovable(axis, wallX, wallY) then
                    self:ensureWallConsidered(axis, wallX, wallY)
                end
            end)

            y = y + 1
        end
    end

    function world.ensureColumnsComplete(self, x1)
        local x2 = x1 + worldMinColumns

        while self.lastColumn <= x2 do
            self:ensureColumnComplete(self.lastColumn)
            self.lastColumn = self.lastColumn + 1
        end

        -- TODO: Should only be called if something actually changed
        self.changedEvent:signal()
    end

    function world.checkMove(self, x, y, direction)
        local axis = wallAxes[direction]
        local wallOffset = offsetWalls[direction]
        local wallX = x + wallOffset[1]
        local wallY = y + wallOffset[2]

        -- Note: This assumes that the wall has been considered already
        if self.walls:get(axis, wallX, wallY) then
            return false
        end

        return true
    end

    function world.reset(self)
        self.lastColumn = 0
        self.walls:clear()
        self.squares:clear()

        self.squaresUsed:clear()
        self.wallsAvailable:clear()
        self.wallsConsidered:clear()

        self:annexSquare(0, 0)
        self:ensureColumnsComplete(0)
    end

    return world
end

Manager = { }

function Manager.new(world, player, ender)
    local manager = Entity.new()
    local managerLevelMovePeriodDelta = 25
    local managerLevelLength = 50
    local level = 0
    local score = 0
    local ended = false

    -- Events
    manager.scoreChangedEvent = Event.new()
    manager.levelChangedEvent = Event.new()
    manager.endedEvent = Event.new()

    player.movedEvent:addHandler(function(x, y)
        local lastScore = score

        score = max(score, x)

        if score ~= lastScore then
            manager.scoreChangedEvent:signal(score)
        end
    end)

    function manager.update(self, ms)
        if not ended and ender.x >= player.x then
            self.endedEvent:signal()
            ended = true
        end

        -- Check for level up
        local lastLevel = level

        level = max(level, floor(player.x / managerLevelLength))

        if level ~= lastLevel then
            ender:setMovePeriod(max(100, enderInitialMovePeriod - level * managerLevelMovePeriodDelta))
            self.levelChangedEvent:signal(level)
        end
    end

    function manager.getScore(self)
        return score
    end

    function manager.reset(self)
        level = 0
        score = 0
        ended = false
    end

    return manager
end

Info = { }

function Info.new(manager)
    local info = Entity.new()
    local levelLabel = Element.Text.new(nil, -200, 90, 1, textWidth, textHeight)
    local scoreLabel = Element.Text.new(nil, 200, 90, 1, textWidth, textHeight)
    local titleElement = Element.Text.new(title, 0, 150, 1, textWidth * 2, textHeight * 2)

    titleElement.alignment = "center"

    local levelBuffer = StringBuffer.new()
    local scoreBuffer = StringBuffer.new()

    levelLabel.buffer = levelBuffer
    scoreLabel.buffer = scoreBuffer
    scoreLabel.alignment = "right"

    local function emphasizeElement(element)
        local text = element.buffer.text
        local x = element.x + #text / 2 * textWidth

        if element.alignment == "right" then
            x = element.x - #text / 2 * textWidth
        end

        local entity = Entity.new(x, element.y + textHeight / 2)

        entity.width = textWidth
        entity.height = textHeight

        entity.elements:add(Element.Text.new(text, -#text / 2, -0.5))
        info:addChild(Ghost.new(entity, 350, 1.5))
    end

    local function updateLevel(level)
        levelBuffer:clear()
        levelBuffer:append(String.format("Level: %d", level))
    end

    local function updateScore(score)
        scoreBuffer:clear()
        scoreBuffer:append(String.format("Score: %d", score))
    end

    manager.levelChangedEvent:addHandler(function(level)
        updateLevel(level)

        -- Special effect on level change
        emphasizeElement(levelLabel)
        Audio.play("Sounds/LevelUp.wav", 0.7)
    end)

    manager.scoreChangedEvent:addHandler(updateScore)

    -- Special effect on end
    manager.endedEvent:addHandler(function()
        emphasizeElement(scoreLabel)
        Audio.play("Sounds/Burn.wav", 0.5)
    end)

    function info.reset(self)
        self.elements:clear()
        self.elements:add(levelLabel)
        self.elements:add(scoreLabel)
        self.elements:add(titleElement)

        updateLevel(0)
        updateScore(0)
    end

    return info
end

Display = { }

local displayColorPlayer = Color.new(1, 1, 1)
local displayColorEnder = Color.new(1, 0, 0, 0.8)
local displayColorBackground = Color.new(0.3, 0.3, 0.3)

function Display.new(world, player, ender, manager)
    local squareSize = 20
    local wallSizeRelative = 0.125
    local playerSizeRelative = 0.5

    local rows = world.y2 - world.y1 + 1
    local columns = 640 / squareSize

    local colorInvisible = Color.new(1, 1, 1, 0)
    local colorVisible = Color.new(1, 1, 1, 1)

    -- Create the display entity
    local display = Entity.new((-columns / 2 + 0.5) * squareSize, (-rows / 2 + 0.5) * squareSize, 1, squareSize, squareSize)

    -- Events
    display.viewportChangedEvent = Event.new()

    -- Viewport
    local vx1 = 0
    local vy1 = world.y1

    function display.getViewport(self)
        return vx1, columns
    end

    -- Walls
    local wallsEntity = Entity.new(0, 0, -1)

    local function forEachWall(f)
        local x = -1

        while x < columns do
            local y = -1

            while y < rows do
                if x > -1 then
                    f(1, x, y)
                end

                if y > -1 then
                    f(2, x, y)
                end
                
                y = y + 1
            end

            x = x + 1
        end
    end

    -- Wall elements
    local walls = { { }, { } }

    forEachWall(function(axis, x, y)
        if walls[axis][x] == nil then
            walls[axis][x] = { }
        end

        if axis == 1 then
            walls[1][x][y] = wallsEntity.elements:add(Element.Image.new("Images/Blank.png", x, y + 0.5, 0, 1, wallSizeRelative))
        else
            walls[2][x][y] = wallsEntity.elements:add(Element.Image.new("Images/Blank.png", x + 0.5, y, 0, wallSizeRelative, 1))
        end
    end)

    -- TODO: Should allow fast column swapping, shouldn't re-check everything if possible
    local function updateWalls()
        forEachWall(function(axis, x, y)
            local wallElement = walls[axis][x][y]

            if world.walls:get(axis, x + vx1, y + vy1) then
                wallElement.color = colorVisible
            else
                wallElement.color = colorInvisible
            end
        end)
    end

    world.changedEvent:addHandler(updateWalls)

    -- Ender
    local enderEntity = Entity.new(0, rows / 2 - 0.5, 2)
    local enderElement = enderEntity.elements:add(Element.Image.new("Images/Blank.png", 0, 0, 10, 1, rows))
    local enderAudibleDistance = 40
    local function updateEnder()
        enderEntity.x = ender.x - vx1
    end

    enderElement.color = displayColorEnder
    ender.movedEvent:addHandler(function()
        updateEnder()

        local volume = 1 - abs(player.x - ender.x) / enderAudibleDistance

        if volume > 0.05 then
            Audio.play("Sounds/Ender.wav", volume)
        end
    end)

    -- Player element
    local playerEntity = Entity.new(0 ,0, 1)
    local playerElement = playerEntity.elements:add(Element.Image.new("Images/Blank.png", 0, 0, 0, playerSizeRelative, playerSizeRelative))
    local centerThreshold = floor(columns / 12)
    local function playerMoved(x, y)
        -- Center the view
        local center = floor(columns / 2)
        local centerX = center + vx1
        local viewportChanged = false

        if centerX - player.x > centerThreshold then
            vx1 = centerThreshold + player.x - center
            viewportChanged = true
        elseif centerX - player.x < -centerThreshold then
            vx1 = -centerThreshold + player.x - center
            viewportChanged = true
        end

        if viewportChanged then
            updateWalls()
            updateEnder()
            display.viewportChangedEvent:signal(display:getViewport())
        end

        -- Update the player element
        playerEntity.x = x - vx1
        playerEntity.y = y - vy1
    end

    player.movedEvent:addHandler(playerMoved)

    -- Background element
    local backgroundEntity = Entity.new(-0.5, -0.5, -5, columns, rows)
    local backgroundElement = backgroundEntity.elements:add(Element.Image.new("Images/Blank.png", 0.5, 0.5, -1))

    backgroundElement.color = displayColorBackground

    local function updateBackground(vx1, columns)
        backgroundEntity.x = max(0, -vx1) - 0.5
        backgroundEntity.width = min(columns, columns + vx1)
    end

    display.viewportChangedEvent:addHandler(updateBackground)

    -- End effect
    local colorPlayerNormal = Color.new(1, 1, 1)
    local colorPlayerEnded = Color.new(1, 0, 0)

    manager.endedEvent:addHandler(function()
        display:removeChild(playerEntity)
        playerEntity.color = colorPlayerEnded
        display:addChild(Ghost.new(playerEntity, 2000, 2))
        playerEntity.color = colorPlayerNormal
    end)

    function display.reset(self)
        self:clearChildren()
        self:addChild(wallsEntity)
        display:addChild(enderEntity)
        display:addChild(playerEntity)
        display:addChild(backgroundEntity)

        vx1 = 0
        vy1 = world.y1
        updateWalls()
        playerMoved(player.x, player.y)
        updateBackground(self:getViewport())
    end

    return display
end

Summary = { }

function Summary.new(world, player, ender, manager, display)
    local summary = Entity.new(-300, -200, 1, 1, 30)
    local lineWidth = 2
    local scale = 0.5

    -- Ender line
    local enderElement = summary.elements:add(Element.Image.new("Images/Blank.png", 0, 0.5, 1, lineWidth))
    local function updateEnder()
        enderElement.x = ender.x * scale
    end

    enderElement.color = displayColorEnder
    ender.movedEvent:addHandler(updateEnder)

    -- Player line
    local playerElement = summary.elements:add(Element.Image.new("Images/Blank.png", 0, 0.5, 1, lineWidth))
    local function updatePlayer()
        playerElement.x = player.x * scale
    end

    playerElement.color = displayColorPlayer
    player.movedEvent:addHandler(updatePlayer)

    -- Viewport
    local viewportElement = summary.elements:add(Element.Image.new("Images/Blank.png", 0, 0.5, 0, lineWidth))
    local function updateViewport(x, width)
        viewportElement.x = (x + width / 2) * scale
        viewportElement.width = width * scale
    end

    viewportElement.color = displayColorBackground
    display.viewportChangedEvent:addHandler(updateViewport)

    function summary.reset(self)
        updateEnder()
        updatePlayer()
        updateViewport(display:getViewport())
    end

    return summary
end

GameLayer = { }

function GameLayer.new(ended)
    local layer = Layer.new(33)
    local paused = false

    local function createDirectionHandler(directionName)
        return function(self, pressed)
            if not paused then
                self.player:directionPressed(directionValues[directionName], pressed)
            end
        end
    end

    local function endGame()
        Layers.pop()
        ended(layer.manager:getScore())
    end

    local eventHandlers = {
        left  = createDirectionHandler("left"),
        right = createDirectionHandler("right"),
        up    = createDirectionHandler("up"),
        down  = createDirectionHandler("down"),

        activate = function(self, pressed)
            if pressed and paused then
                endGame()
            end
            end,

        cancel = function(self, pressed)
            if pressed then
                endGame()
            end
            end,
    }

    input:setupEventHandlers(layer)

    layer.world = layer:addChild(World.new())
    layer.player = layer.world:addChild(Player.new(layer.world))
    layer.ender = layer.world:addChild(Ender.new())

    layer.manager = layer:addChild(Manager.new(layer.world, layer.player, layer.ender))

    layer.manager.endedEvent:addHandler(function()
        layer.player:setActive(false)
        paused = true
    end)

    layer.info = layer:addChild(Info.new(layer.manager))
    layer.display = layer:addChild(Display.new(layer.world, layer.player, layer.ender, layer.manager))
    layer.summary = layer:addChild(Summary.new(layer.world, layer.player, layer.ender, layer.manager, layer.display))
    
    function layer.inputReceived(self, event, pressed)
        local eventHandler = eventHandlers[event]
        local handled = false
    
        if isFunction(eventHandler) then
            eventHandler(self, pressed)
            handled = true
        end
    
        return handled
    end

    function layer.reset(self)
        layer:clearAudio()
        paused = false
        self.world:reset()
        self.player:reset()
        self.ender:reset()
        self.manager:reset()
        self.info:reset()
        self.display:reset()
        self.summary:reset()
    end

    return layer
end

-- Start the game
require("Menus/MainMenu.lua")

Video.setTitle(title)
Layers.push(MainMenu.new())

